#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iksemel.h>


#ifdef WIN32
#include <float.h>
#define HUGE_VAL _FPCLASS_PINF
#else
/* Informs glibc to define NAN, etc */
#define __USE_ISOC99 1
/* Just to define NAN, shouldn't need -lm */
#include <math.h>
#endif

#include "SDL_datadir.h"
#include "SDL_conf.h"

#define COMPACT_RATE 1024
#define CONF_VERSION 0.3
#define CONF_VERSTR "0.3"
#define CONF_DELIM '/'


#ifndef NAN
#ifndef WIN32
#warning "NAN not defined, trying to make do"
#endif/*WIN32*/
#ifdef HUGE_VAL
#define NAN (HUGE_VAL/HUGE_VAL)
#else
#error "NAN not defined!"
#endif/*HUGE_VAL*/
#endif/*!NAN*/

const float BAD_FLOAT=NAN;

/* Global path to dynamic data files */
static char gpath[512]={'\0'};
/* Global path to static data files */
static char spath[512]={'\0'};
/* Global flags -- autosave, etc */
static int  gflags=0;
/* Whether global variables have been initialized */
static int  init=0;
/* When <=0, recompact the tree. */
static int  docompact=COMPACT_RATE;
/* The root of the data tree */
static iks *tree=NULL;

/* Just like strncpy, except safe null-termination is guaranteed */
static void strnncpy(char *dest, const char *src, size_t len)
{
  if(len<=0) return;
  strncpy(dest,src,len-1);  
  dest[len-1]='\0';
}

const char *SDL_ConfStaticPath()
{
  if(!init) return(NULL);
  return(spath);
}

/**
 * Recreates the whole tree to eliminate spaceholes from frequent updating.
 */
int SDL_ConfCompact()
{
  iks *tmp,*old;

  if(!init) return(-1);
  old=tree;
  tmp=iks_copy(tree);
  if(tmp==NULL) return(-1);

  tree=tmp;
  iks_delete(old);
  return(0);
}
/**
 * Repacks the tree after COMPACT_RATE tree modifications
 */
void SDL_ConfAutopack()
{
  if(docompact<=0)
  {
    SDL_ConfCompact();
    docompact=COMPACT_RATE;
  }
}

/**
 * Takes a string of the format "foo/bar/etc/stuff", and seeks
 * to root -> foo -> bar -> etc -> stuff to the tree and returns
 * the node 'stuff'.  If create is zero, it will fail if these
 * branches do not exist;  if create is nonzero, it will attempt
 * to create branches as it goes.
 */
void *SDL_ConfSeekNext(void *curpos,const char *str, int create)
{
  char buf[512];
//  const char *attr;
  iks *iks_next;
  const char *next=strchr(str,CONF_DELIM);

  if(curpos==NULL) curpos=(void *)tree;
  if(!init) return(NULL);

  while(next!=NULL)
  {
    strnncpy(buf,str,(next-str)+1);
    iks_next=iks_find((iks *)curpos,buf);

    if(iks_next==NULL)
    {
      int e;
      if(!create) return(NULL);

      iks_next = iks_insert((iks *)curpos,buf);
      if(iks_next==NULL) return(NULL);
    }

    str=next+1;
    next=strchr(str,CONF_DELIM);
    curpos=(void *)iks_next;
  }

  iks_next=iks_find((iks *)curpos,str);
  if(iks_next==NULL)
  {
    if(!create) return(NULL);

    iks_next = iks_insert((iks *)curpos,str);
  }

  return((void *)iks_next);
}

int SDL_ConfTagRemove(void *root)
{
  if(!init) return(-1);

  if(root==tree)
  {
    fprintf(stderr,"Will NOT remove the root of the data tree!\n");
    return(-1);
  }

  iks_hide((iks *)root);
  return(0);
}


/**
 * Attempts to initialize the global configuration system.
 * Searches dynamic paths for an available config directory,
 * loads the config file from if present or creates a new one if not.
 * You do NOT need to specify where the files go, just give a unique
 * application name.
 */
int SDL_ConfLoad(const char *app, int flags)
{
  const char *ver;
  char buf[512]={'\0'};
  if(init) return(0);

  if(SDL_DirDynamic(gpath,app,512,flags&CONF_CREATE)==NULL)
    return(-1);

  fprintf(stderr,"Have conf dir:\t'%s'\n",gpath);

  if(SDL_DirStatic(spath,app,512)<0)
    fprintf(stderr,"Warning, static path could not be found\n");
  else
    fprintf(stderr,"Have static dir:\t'%s'\n",spath);

  sprintf(buf,"%spersistent.xml",gpath);

  if(iks_load(buf,&tree)==IKS_OK)
  {
    float filever;
    if(strcmp("sdlconf",iks_name(tree))!=0)
    {
      fprintf(stderr,"Invalid config file root '%s'\n",iks_name(tree));
      iks_delete(tree);
      tree=NULL;
      return(-1);
    }

    filever=SDL_ConfGetFloatDefault(tree,"version",NAN);
    if(filever == NAN)
      fprintf(stderr,"Warning -- version unspecified\n");
    else if((filever - CONF_VERSION)>0.01)
    {
      fprintf(stderr,"File is newer than SDL_conf!\n");
      iks_delete(tree);
      tree=NULL;
      return(-1);
    }

    fprintf(stderr,"Loaded persistent data tree\n");
    init=1;
    gflags=flags;
    return(0);
  }
  else if(flags&CONF_CREATE)
  {
    tree=iks_new("sdlconf");
    iks_insert_attrib(tree,"version",CONF_VERSTR);
    fprintf(stderr,"Created new persistent data tree\n");
    gflags=flags;
    init=1;
    return(0);
  }
  else
  {
    fprintf(stderr,"Unable to load or create persistent data\n");
    return(-1);
  }
}

int SDL_ConfSave()
{
  char buf[512];
  if(!init) return(-1);

  sprintf(buf,"%spersistent.xml",gpath);
  if(iks_save(buf,tree)!=IKS_OK)
    return(-1);
  fprintf(stderr,"Saved persistent data tree to %s\n",buf);
  return(0);
}

/**
 * Frees the data tree and sets SDL_conf back to uninitialized state.
 * If SDL_conf was initialized with CONF_AUTOSAVE, it automatically
 * resaves the data tree to file before doing this.  You can run
 * SDL_ConfClose() in your atexit() to make this automatic on exit.
 */
void SDL_ConfClose()
{
  if(!init) return;

  if(gflags&CONF_AUTOSAVE)
    if(SDL_ConfSave()<0)
      fprintf(stderr,"Warning!  Autosave failed!\n");

  iks_delete(tree);
  tree=NULL;
  init=0;
}

/*! \brief Sets an attribute */
int SDL_ConfSet(void *root, const char *att,const char *val)
{
  if(root == NULL) root=(void *)tree;
  if(iks_insert_attrib((iks *)root,att,val)==NULL)
    return(-1);
  else
    return(0);
}

const char *SDL_ConfGet(void *root, const char *att)
{
  if(root == NULL) root=(void *)tree;

  return(iks_find_attrib((iks *)root,att));
}

int SDL_ConfSetLong(void *root, const char *att, long val)
{
  char lbuf[32];
  sprintf(lbuf,"%ld",val);
  return(SDL_ConfSet(root,att,lbuf));
}

long SDL_ConfGetLongDefault(void *root, const char *att, long def)
{
  const char *cbuf=SDL_ConfGet(root,att);
  if(cbuf==NULL)
    return(def);
  else
  {
    long r=strtol(cbuf,NULL,0);
    return(r);
  } 
}

int SDL_ConfSetFloat(void *root, const char *att, float val)
{
  char buf[64];
  sprintf(buf,"%f",val);
  return(SDL_ConfSet(root,att,buf));
}

float SDL_ConfGetFloatDefault(void *root, const char *att, float def)
{
  float fl;
  const char *f=SDL_ConfGet(root,att);

  if(f==NULL)
    return(def);
  else if(sscanf(f,"%f",&fl)!=1)
    return(def);
  else
    return(fl);
}

int SDL_ConfSetBool(void *root, const char *name, int val)
{
  if(val)
    return(SDL_ConfSet(root,name,TRUE_STR));
  else
    return(SDL_ConfSet(root,name,FALSE_STR));
}

int SDL_ConfGetBoolDefault(void *root, const char *name, int def)
{
  const char *bstr=SDL_ConfGet(root,name);
  if(bstr==NULL)
    return(def!=0);
  else if(strcmp(bstr,TRUE_STR)==0)
    return(1);
  else
    return(0);
}

/**
 * Returns nonzero if the named path exists in the persistent data tree, 
 * zero if not.
 */
int SDL_ConfTagExists(void *root,const char *name)
{
  if(!init) return(0);
  return(SDL_ConfSeekNext(root,name,0)!=NULL);
}

int SDL_ConfSetDefault(void *root, const char *var, const char *val)
{
  if(SDL_ConfGet(root,var)!=NULL)
    return(SDL_ConfSet(root,var,val));

  return(-1);
}
